searchNodes=[{"ref":"Telemetry.Metrics.html","title":"Telemetry.Metrics","module":"Telemetry.Metrics","type":"module","doc":"Common interface for defining metrics based on :telemetry events. Metrics are aggregations of Telemetry events with specific name, providing a view of the system&#39;s behaviour over time. For example, to build a sum of HTTP request payload size received by your system, you could define a following metric: sum(&quot;http.request.payload_size&quot;) This definition means that the metric is based on [:http, :request] events, and it should sum up values under :payload_size key in events&#39; measurements. Telemetry.Metrics also supports breaking down the metric values by tags - this means that there will be a distinct metric for each unique set of selected tags found in event metadata: sum(&quot;http.request.payload_size&quot;, tags: [:host, :method]) The above definiton means that we want to keep track of the sum, but for each unique pair of request host and method (assuming that :host and :method keys are present in event&#39;s metadata). There are four metric types provided by Telemetry.Metrics: counter/2 which counts the total number of emitted events sum/2 which keeps track of the sum of selected measurement last_value/2 holding the value of the selected measurement from the most recent event distribution/2 which builds a histogram of selected measurement Note that these metric definitions by itself are not enough, as they only provide the specification of what is the expected end-result. The job of subscribing to events and building the actual metrics is a responsibility of reporters (described in the &quot;Reporters&quot; section). Metric definitions Metric definition is a data structure describing the metric - its name, type, name of the events aggregated by the metric, etc. The structure of metric definition is relevant only to authors of reporters. Metric definitions are created using one of the four functions: counter/2, sum/2, last_value/2 and distribution/2. Each of those functions returns a definition of metric of the corresponding type. The first argument to all these functions is the metric name. Metric name can be provided as a string (e.g. &quot;http.request.latency&quot;) or a list of atoms ([:http, :request, :latency]). If not overriden in the metric options, metric name also determines the name of Telemetry event and measurement used to produce metric values. In the &quot;http.request.latency&quot; example, the source event name is [:http, :request] and metric values are drawn from :latency measurement. Note: do not use data from external sources as metric or event names! Since they are converted to atoms, your application becomes vulnerable to atom leakage and might run out of memory. The second argument is a list of options. Below is the description of the options common to all metric types: :event_name - the source event name. Can be represented either as a string (e.g. &quot;http.request&quot;) or a list of atoms ([:http, :request]). By default the event name is all but the last segment of the metric name. :measurement - the event measurement used as a source of a metric values. By default it is the last segment of the metric name. It can be either an arbitrary term, a key in the event&#39;s measurements map, or a function accepting the whole measurements map and returning the actual value to be used. :tags - a subset of metadata keys by which aggregations will be broken down. Defaults to an empty list. :tag_values - a function that receives the metadata and returns a map with the tags as keys and their respective values. Defaults to returning the metadata itself. :description - human-readable description of the metric. Might be used by reporters for documentation purposes. Defaults to nil. :unit - an atom describing the unit of selected measurement or a tuple indicating that a measurement should be converted from one unit to another before a metric is updated. Currently, only time unit conversions are supported. For example, setting this option to {:native, :millisecond} means that the measurements are provided in the :native time unit (you can read more about it in the documentation for System.convert_time_unit/3), but a metric should have its values in milliseconds. Both elements of the conversion tuple need to be of type time_unit/0. Reporters Reporters take metric definitions as an input, subscribe to relevant events and update the metrics when the events are emitted. Updating the metric might involve publishing the metrics periodically, or on demand, to external systems. Telemetry.Metrics defines only how metrics of particular type should behave and reporters should provide actual implementation for these aggregations. Telemetry.Metrics package does not include any reporter itself."},{"ref":"Telemetry.Metrics.html#counter/2","title":"Telemetry.Metrics.counter/2","module":"Telemetry.Metrics","type":"function","doc":"Returns a definition of counter metric. Counter metric keeps track of the total number of specific events emitted. Note that for the counter metric it doesn&#39;t matter what measurement is selected, as it is ignored by reporters anyway. See the &quot;Metric definitions&quot; section in the top-level documentation of this module for more information. Example counter( &quot;http.request.count&quot;, tags: [:controller, :action] )"},{"ref":"Telemetry.Metrics.html#distribution/2","title":"Telemetry.Metrics.distribution/2","module":"Telemetry.Metrics","type":"function","doc":"Returns a definition of distribution metric. Distribution metric builds a histogram of selected measurement&#39;s values. Because of that, it is required that you specify the histograms buckets via :buckets option. For example, given buckets: [0, 100, 200], the distribution metric produces four values: number of measurements less than or equal to 0 number of measurements greater than 0 and less than or equal to 100 number of measurements greater than 100 and less than or equal to 200 number of measurements greater than 200 See the &quot;Metric definitions&quot; section in the top-level documentation of this module for more information. Example distribution( &quot;http.request.latency&quot;, buckets: [100, 200, 300], tags: [:controller, :action], )"},{"ref":"Telemetry.Metrics.html#last_value/2","title":"Telemetry.Metrics.last_value/2","module":"Telemetry.Metrics","type":"function","doc":"Returns a definition of last value metric. Last value keeps track of the selected measurement found in the most recent event. See the &quot;Metric definitions&quot; section in the top-level documentation of this module for more information. Example last_value( &quot;vm.memory.total&quot;, description: &quot;Total amount of memory allocated by the Erlang VM&quot;, unit: :byte )"},{"ref":"Telemetry.Metrics.html#sum/2","title":"Telemetry.Metrics.sum/2","module":"Telemetry.Metrics","type":"function","doc":"Returns a definition of sum metric. Sum metric keeps track of the sum of selected measurement&#39;s values carried by specific events. See the &quot;Metric definitions&quot; section in the top-level documentation of this module for more information. Example sum( &quot;user.session_count&quot;, event_name: &quot;user.session_count&quot;, measurement: :delta, tags: [:role] )"},{"ref":"Telemetry.Metrics.html#t:counter_options/0","title":"Telemetry.Metrics.counter_options/0","module":"Telemetry.Metrics","type":"type","doc":""},{"ref":"Telemetry.Metrics.html#t:description/0","title":"Telemetry.Metrics.description/0","module":"Telemetry.Metrics","type":"type","doc":""},{"ref":"Telemetry.Metrics.html#t:distribution_options/0","title":"Telemetry.Metrics.distribution_options/0","module":"Telemetry.Metrics","type":"type","doc":""},{"ref":"Telemetry.Metrics.html#t:last_value_options/0","title":"Telemetry.Metrics.last_value_options/0","module":"Telemetry.Metrics","type":"type","doc":""},{"ref":"Telemetry.Metrics.html#t:measurement/0","title":"Telemetry.Metrics.measurement/0","module":"Telemetry.Metrics","type":"type","doc":""},{"ref":"Telemetry.Metrics.html#t:metric_name/0","title":"Telemetry.Metrics.metric_name/0","module":"Telemetry.Metrics","type":"type","doc":"The name of the metric, either as string or a list of atoms."},{"ref":"Telemetry.Metrics.html#t:metric_option/0","title":"Telemetry.Metrics.metric_option/0","module":"Telemetry.Metrics","type":"type","doc":""},{"ref":"Telemetry.Metrics.html#t:normalized_metric_name/0","title":"Telemetry.Metrics.normalized_metric_name/0","module":"Telemetry.Metrics","type":"type","doc":"The name of the metric represented as a list of atoms."},{"ref":"Telemetry.Metrics.html#t:sum_options/0","title":"Telemetry.Metrics.sum_options/0","module":"Telemetry.Metrics","type":"type","doc":""},{"ref":"Telemetry.Metrics.html#t:t/0","title":"Telemetry.Metrics.t/0","module":"Telemetry.Metrics","type":"type","doc":"Common fields for metric specifications Reporters should assume that these fields are present in all metric specifications."},{"ref":"Telemetry.Metrics.html#t:tag/0","title":"Telemetry.Metrics.tag/0","module":"Telemetry.Metrics","type":"type","doc":""},{"ref":"Telemetry.Metrics.html#t:tag_values/0","title":"Telemetry.Metrics.tag_values/0","module":"Telemetry.Metrics","type":"type","doc":""},{"ref":"Telemetry.Metrics.html#t:tags/0","title":"Telemetry.Metrics.tags/0","module":"Telemetry.Metrics","type":"type","doc":""},{"ref":"Telemetry.Metrics.html#t:time_unit/0","title":"Telemetry.Metrics.time_unit/0","module":"Telemetry.Metrics","type":"type","doc":""},{"ref":"Telemetry.Metrics.html#t:unit/0","title":"Telemetry.Metrics.unit/0","module":"Telemetry.Metrics","type":"type","doc":""},{"ref":"Telemetry.Metrics.html#t:unit_conversion/0","title":"Telemetry.Metrics.unit_conversion/0","module":"Telemetry.Metrics","type":"type","doc":""},{"ref":"Telemetry.Metrics.Counter.html","title":"Telemetry.Metrics.Counter","module":"Telemetry.Metrics.Counter","type":"module","doc":"Defines a specification of counter metric."},{"ref":"Telemetry.Metrics.Counter.html#t:t/0","title":"Telemetry.Metrics.Counter.t/0","module":"Telemetry.Metrics.Counter","type":"type","doc":""},{"ref":"Telemetry.Metrics.Distribution.html","title":"Telemetry.Metrics.Distribution","module":"Telemetry.Metrics.Distribution","type":"module","doc":"Defines a specification of distribution metric."},{"ref":"Telemetry.Metrics.Distribution.html#t:buckets/0","title":"Telemetry.Metrics.Distribution.buckets/0","module":"Telemetry.Metrics.Distribution","type":"type","doc":"Distribution metric bucket boundaries. Bucket boundaries are represented by a non-empty list of increasing numbers. Examples [0, 100, 200, 300] # Buckets: [-inf, 0], [0, 100], [100, 200], [200, 300], [300, +inf] [99.9] # Buckets: [-inf, 99.9], [99.9, +inf]"},{"ref":"Telemetry.Metrics.Distribution.html#t:t/0","title":"Telemetry.Metrics.Distribution.t/0","module":"Telemetry.Metrics.Distribution","type":"type","doc":""},{"ref":"Telemetry.Metrics.LastValue.html","title":"Telemetry.Metrics.LastValue","module":"Telemetry.Metrics.LastValue","type":"module","doc":"Defines a specification of last value metric."},{"ref":"Telemetry.Metrics.LastValue.html#t:t/0","title":"Telemetry.Metrics.LastValue.t/0","module":"Telemetry.Metrics.LastValue","type":"type","doc":""},{"ref":"Telemetry.Metrics.Sum.html","title":"Telemetry.Metrics.Sum","module":"Telemetry.Metrics.Sum","type":"module","doc":"Defines a specification of sum metric."},{"ref":"Telemetry.Metrics.Sum.html#t:t/0","title":"Telemetry.Metrics.Sum.t/0","module":"Telemetry.Metrics.Sum","type":"type","doc":""},{"ref":"overview.html","title":"Overview","module":"overview","type":"extras","doc":"Overview Telemetry.Metrics provides a common interface for defining metrics based on :telemetry events. While a single event means that a thing happened (e.g. an HTTP request was sent or a DB query returned a result), a metric is an aggregation of those events over time. For instance, we could count the number of HTTP requests, or keep track of the sum of payload sizes returned by DB queries. To give a more concrete example, imagine that somewhere in your code you have a function which sends an HTTP request, measures the time it took to get a response, and emits an event with the information: :telemetry.execute([:http, :request, :done], %{duration: duration}) You could define a counter metric, which counts how many HTTP requests were completed: Telemetry.Metrics.counter(&quot;http.request.done.count&quot;) or you could use a distribution metric to see how many queries were completed in particular time buckets: Telemetry.Metrics.distribution(&quot;http.request.done.duration&quot;, buckets: [100, 200, 300]) There are a couple more metric types, and they are all described in detail in the &quot;Metric types&quot; section. Because of the fact that metrics are based only on events being emitted, you can easily create metrics from events published by the libraries you&#39;re using in your project. But metric definitions on their own are not enough - aggregated metrics need to be sent somewhere, so that you can inspect how your system behaves - and that&#39;s what reporters are for."},{"ref":"overview.html#reporters","title":"Overview - Reporters","module":"overview","type":"extras","doc":"Reporters are responsible for publishing metrics to some system where they can be inspected. For example, there could be a reporter pushing metrics to StatsD, some time-series database, or exposing a HTTP endpoint for Prometheus to scrape. Under the hood, reporter needs to attach event handlers to relevant events and extract specific measurement. This information is included in the metric definitions. Note that Telemetry.Metrics package doesn&#39;t provide any reporter itself."},{"ref":"overview.html#metric-definitions","title":"Overview - Metric definitions","module":"overview","type":"extras","doc":"counter/2, sum/2, last_value/2 and distribution/2 functions all return metric defintions. The most basic metric definition looks like this sum(&quot;http.request.payload_size&quot;) The first argument to the metric definition function is a metric name - this is what the reporter will use to identify this metric when publishing it. The metric name also determines what event and measurement should be used to produce metric values: [:http , :request] :payload_size &lt;-- event name --&gt; &lt;-- measurement --&gt; that is, by default, all but last segments of the metric name determine the event name, and the last segment determines the measurement. If you wish to use a different event name or measurement, they can be overriden using :event_name and :measurement options respectively (you can read more about them in the &quot;Shared options&quot; section in the docs for Telemetry.Metrics module)"},{"ref":"overview.html#metric-types","title":"Overview - Metric types","module":"overview","type":"extras","doc":"Telemetry.Metrics defines four basic metric types: a counter simply counts the number of emitted events, regardless of measurements included in the events. Since the measurement does not matter in case of a counter, we recommend using count as a measurement, e.g. &quot;http.request.count&quot; a last_value metric holds the value of a selected measurement found in the most recent event a sum adds up the values of a selected measurement in all the events a distribution keeps track of the histogram of the selected measurement, i.e. how many measurements fall into defined buckets. Histogram allows to compute useful statistics about the data, like percentiles, minimum, or maximum. For example, given boundaries [0, 100, 200], the distribution metric produces four values: number of measurements less than or equal to 0 number of measurements greater than 0 and less than or equal to 100 number of measurements greater than 100 and less than or equal to 200 number of measurements greater than 200 Note that not all metric types are supported by all monitoring solutions. However, often they support some variation of a particular type. For example, StatsD doesn&#39;t have a distribution metric as defined here built-in, but it provides a &quot;timer&quot; metric which allows to keep track of the percentiles, maximum, etc. That is fine as long as the reporter properly documents the differences between the expected and actual behaviour. It&#39;s also possible that a reporter library provides its own, specialized function for building metric definitions, for metric types specific to the system it publishes metrics to."},{"ref":"overview.html#breaking-down-metric-values-by-tags","title":"Overview - Breaking down metric values by tags","module":"overview","type":"extras","doc":"Sometimes it&#39;s not enough to have a global overview of all HTTP requests received or all DB queries made. It&#39;s often more helpful to break down this data, for example, we might want to have separate metric values for each unique database table and operation name (select, insert etc.) to see how these particular queries behave. This is where tagging comes into play. All metric definitions accept a :tags option: count(&quot;db.query.count&quot;, tags: [:table, :operation]) The above definition means that we want to keep track of the number of queries, but we want a separate counter for each unique pair of table and operation. Tag values are fetched from event metadata - this means that in this example, [:db, :query] events need to include :table and :operation keys in their payload: :telemetry.execute([:db, :query], %{duration: 198}, %{table: &quot;users&quot;, operation: &quot;insert&quot;}) :telemetry.execute([:db, :query], %{duration: 112}, %{table: &quot;users&quot;, operation: &quot;select&quot;}) :telemetry.execute([:db, :query], %{duration: 201}, %{table: &quot;sessions&quot;, operation: &quot;insert&quot;}) :telemetry.execute([:db, :query], %{duration: 212}, %{table: &quot;sessions&quot;, operation: &quot;insert&quot;}) The result of aggregating the events above looks like this: tableoperationcount usersinsert1 usersselect1 sessionsinsert2 The approach where we create a separate metric for some unique set of properties is called a multi-dimensional data model. Transforming event metadata for tagging Finally, sometimes there is a need to modify event metadata before it&#39;s used for tagging. Each metric definition accepts a function in :tag_values option which transforms the metadata into desired shape. Note that this function is called for each event, so it&#39;s important to keep it fast if the rate of events is high."},{"ref":"overview.html#converting-units","title":"Overview - Converting units","module":"overview","type":"extras","doc":"It might happen that the unit of measurement we&#39;re tracking is not the desirable unit for the metric values, e.g. events are emitted by a 3rd-party library we do not control, or a reporter we&#39;re using requires specific unit of measurement. For these scenarios, each metric definition accepts a :unit option in a form of a tuple: distribution(&quot;http.request.duration&quot;, unit: {from_unit, to_unit}) This means that the measurement will be converted from from_unit to to_unit before being used for updating the metric. Currently, only time conversions are supported, which means that both from_unit and to_unit need to be one of :second, :millisecond, :microsecond, :nanosecond, or :native. For example, to convert HTTP request duration from :native time unit to milliseconds you&#39;d write: distribution(&quot;http.request.duration&quot;, unit: {:native, :millisecond})"},{"ref":"overview.html#vm-metrics","title":"Overview - VM metrics","module":"overview","type":"extras","doc":"Telemetry.Metrics doesn&#39;t have a special treatment for the VM metrics - they need to be based on the events like all other metrics. Telemetry.Poller package exposes a bunch of VM-related metrics via :telemetry events. For example, when you add it to your dependencies, you can create a metric keeping track of total allocated VM memory: last_value(&quot;vm.memory.total&quot;) The last value metric is usually the best fit for VM metrics exposed by the Poller, as the events are emitted periodically and we&#39;re only interested in the most recent measurement. You can read more about available events and measurements in the Telemetry.Poller documentation."},{"ref":"rationale.html","title":"Rationale","module":"rationale","type":"extras","doc":"Rationale The design proposed by Telemetry.Metrics might look controversial - unlike most of the libraries available on the BEAM, it doesn&#39;t aggregate metrics itself, it merely defines what users should expect when using the reporters. If Telemetry.Metrics would aggregate metrics, the way those aggregations work would be imposed on the system where the metrics are published to. For example, counters in StatsD are reset on every flush and can be decremented, whereas counters in Prometheus are monotonically increasing. Telemetry.Metrics doesn&#39;t focus on those details - instead, it describes what the end user, operator, expects to see when using the metric of particular type. This implies that in most cases aggregated metrics won&#39;t be visible inside the BEAM, but in exchange aggregations can be implemented in a way that makes most sense for particular system. Finally, one could also implement an in-VM &quot;reporter&quot; which would aggregate the metrics and expose them inside the BEAM. When there is a need to swap the reporters, and if both reporters are following the metric types specification, then the end result of aggregation is the same, regardless of the backend system in use."},{"ref":"writing_reporters.html","title":"Writing reporters","module":"writing_reporters","type":"extras","doc":"Writing reporters Reporters are a crucial part of Telemetry.Metrics &quot;ecosystem&quot; - without them, metric definitions are merely.. definitions. This guide aims to help in writing the reporter in a proper way. Before writing the reporter for your favourite monitoring system, make sure that one isn&#39;t already available on Hex.pm - it might make sense to contribute and improve the existing solution than starting from scratch. Let&#39;s get started!"},{"ref":"writing_reporters.html#responsibilites","title":"Writing reporters - Responsibilites","module":"writing_reporters","type":"extras","doc":"The reporter has four main responsibilities: it needs to accept a list of metric definitions as input when being started it needs to attach handlers to events contained in these definitions when the events are emitted, it needs to extract the measurement and selected tags, and handle them in a way that makes sense for whathever it chooses to publish to it needs to detach event handlers when it stops or crashes Accepting metric definitions as input This one is quite easy - you need to give your users a way to actually tell you what metrics they want to track. It&#39;s essential to give users an option to provide metric definitions at runtime (e.g. when their application starts). For example, let&#39;s say you&#39;re building a PigeonReporter. If the reporter was process-based, you could provide a start_link/1 function that accepts a list of metric definitions: metrics = [ counter(&quot;...&quot;), last_value(&quot;...&quot;), distribution(&quot;...&quot;) ] PigeonReporter.start_link(metrics) If the reporter doesn&#39;t support metrics of particular type, it log a warning or return an error. Attaching event handlers Event handlers are attached using :telemetry.attach/4 function. To reduce overhead of installing many event handlers, you can install a single handler for multiple metrics based on the same event. Note that handler IDs need to be unique - you can generate completely random blobs of data, or use something that you know needs to be unique anyway, e.g. some combination of reporter name, event name, and something which is different for multiple instances of the same reporter (PID is a good choice if the reporter is process-based): id = {PigeonReporter, metric.event_name, self()} Assuming that metrics is a list of metric definitions based on event, we can attach a handler like this: :telemetry.attach(id, event, &amp;PigeonReporter.handle_event/4, %{metrics: metrics}) Reacting to events There are two parts to event handling - the first one is extracting event measurements and tags, which is the same for all reporters, and the second one is performing logic specific to particular reporter. Let&#39;s implement the basic event handler attached in the previous section: def handle_event(_event_name, measurements, metadata, %{metrics: metrics}) do for metric &lt;- metrics do measurement = extract_measurement(metric, measurements) tags = extract_tags(metric, metadata) # everything else is specific to particular reporter end end As described before, first we extract the measurement and tags, and later perform reporter-specific logic. The implementation of extract_measurement/2 might look as follows: def extract_measurement(metric, measurements) do case metric.measurement do fun when is_function(fun, 1) -&gt; fun.(measurements) key -&gt; measurements[key] end end Since :measurement in the metric definition can be both arbitrary term (to be used as key to fetch the measurement) or a function, we need to handle both cases. Note: Telemetry.Metrics can&#39;t guarantee that the extracted measurement&#39;s value is a number. Each reporter can handle this scenario properly, either by logging a warning, detaching the handler etc. We also need to implement the extract_tags/2 function: def extract_tags(metric, metadata) do tag_values = metric.tag_values.(metadata) for tag &lt;- tags, into: %{} do case Map.fetch(tag_values, tag) do {:ok, value} -&gt; Map.put(tags, tag, value) :error -&gt; Logger.warn(&quot;Tag \#{inspect(tag)} not found in event metadata: \#{inspect(metadata)}&quot;) Map.put(tags, tag, nil) end end end First we need to apply last-minute transformation to the metadata using the :tag_values function. After that, we loop through the list of desired tags and fetch them from transformed metadata - if the particular key is not present in the metadata, we log a warning and assign nil as the tag value. It is very important that the code executed on every event does not fail, as that would cause the handler to be permanently removed and prevent the metrics from being updated. Detaching the handlers on termination To leave the system in a clean state, the reporter should detach the event handlers it installed when it&#39;s being stopped or terminated unexpectedely. This can be done by trapping exists and implementing the terminate callback, or having a dedicated process responsible only for the cleanup (e.g. by using monitors)."},{"ref":"writing_reporters.html#documentation","title":"Writing reporters - Documentation","module":"writing_reporters","type":"extras","doc":"It&#39;s extremely important that reporters document how Telemetry.Metrics metric types, names, and tags are translated to metric types and identifiers in the system they publish metrics to. They should also document if some metric types are not supported at all."},{"ref":"writing_reporters.html#examples","title":"Writing reporters - Examples","module":"writing_reporters","type":"extras","doc":"To our knowledge, there are not many reporters in the wild yet. TelemetryMetricsStatsd is a reporter which might serve as an example when implementing your own."}]